
#include "Arduino.h"
#include "RGB_sprite.h"
#include "masterScreen.h"

//Obj pointer used to fix scope issue with ISR
static masterScreen * volatile cur_screen_obj_ptr;
//which col of the LED Matrix display is being updated
static volatile byte cur_col;

/*
| - Description
|   - used to initialize and create screen,
|   - masterScreen gameScreen;
|   - gameScreen.createScreen();
|
*/
void masterScreen::createScreen()
{
  init();
}


/*
| - Description
|   - Used for testing purposes only, This function prints the screen
|     array
|   - to the Serial monitor in the Arduino IDE. Not meant for use games,
|     will
|   - slow down operation, and errors may occur with LED display
|
*/
void masterScreen::printMasterScreen()
{
  for(int col = 0; col <= 7; col++){
    Serial.print("[");
    for(int row = 0; row <= 7; row++){
      Serial.print(_masterScreen[row][col]);
    }
    Serial.println("]");
  }
  Serial.println("");
}


/*
| - Description
|   - Get the value of the Screen Array of the coordinates entered
|
| - Parameter
|   - col    col, or x value coordinate to be recieved
|   - row    row, or y value coordinate to be recieved
|
*/
byte masterScreen::getScreenValue(byte col, byte row)
{
  return _masterScreen[col][row];
}


/*
| - Description
|   - Modify the coordinates entered with the value enter of the Screen
|     Array
|
| - Parameter
|   - col    col, or x value coordinate to be modified
|   - row    row, or y value coordinate to be modified
|   - value  The value to be placed at those coordinates
|
*/
void masterScreen::modifyScreen(byte col, byte row, byte value)
{
  _masterScreen[col][row] = value;
}


/*
| - Description
|   - takes RGB_sprite sprite_object and updates the _masterScreen[8][8]
|   - Usually used after collision detection and update origin is done to
|   - update that particular sprite
|
| - Parameter
|   - sprite_object   the sprite object to be updated to the
|     _masterScreen[8][8]
|
*/
void masterScreen::updateMasterScreen(RGB_Sprite sprite_object)
{
  byte spriteRow = 0;
  byte spriteCol = 0;
  byte x = sprite_object.origin[0];
  byte y = sprite_object.origin[1];
  byte temp_val = 0;
  for(byte row = y; row < (y + sprite_object.height()) ; row++){
    for(byte col = x; col < (x + sprite_object.width()); col++){
      if((row <= 7) && (col <= 7))
      {
         if(sprite_object.getSpriteValue(spriteCol,spriteRow) != 0){
           temp_val = sprite_object.getSpriteValue(spriteCol,spriteRow);
           _masterScreen[col][row] = temp_val;
         }
        spriteCol++;
      }
    }
    spriteCol = 0;
    spriteRow++;
  }
}

/*
| - Description
|   - takes RGB_sprite sprite_obj[] array and updates the
|     _masterScreen[8][8]
|   - loops the size of the sprite array and calls
|     updateMasterScreen(sprite_obj[i])
|
| - Parameter
|   - num_sprites    number of sprite elements in the array so you can
|     properly loop
|   - sprite_obj[]   the sprite object to be updated to the
|     _masterScreen[8][8]
|
*/
void masterScreen::updateScreen(byte num_sprites, RGB_Sprite sprite_obj[])
{
  for(byte i = 0; i < num_sprites; i++){
    updateMasterScreen(sprite_obj[i]);
  }

}

/*
| - Description
|   - Clears what ever values are on the screen array to zero
|
*/
void masterScreen::clearMasterScreen()
{
  for(byte row = 0; row < 8; row++)
  {
    for(byte col = 0; col < 8; col++)
    {
      _masterScreen[col][row] = 0;
    }
  }
}

/*
| - Description
|   - he sprite_object passed to the function is used to
|   - clear that sprite from the screen
|
| - Parameter
|   - sprite_object
|
*/
void masterScreen::clearSprite(RGB_Sprite sprite_object)
{
  byte x = sprite_object.origin[0];
  byte y = sprite_object.origin[1];
  if( x == 0x80 || y == 0x80)
  {
    //Do nothing, origin not initialized on board
  }else{
    for(byte row = y; row < (y + sprite_object.height()); row++)
    {
      for(byte col = x; col < (x + sprite_object.width()); col++)
      {
        if(row <= 7 && col <= 7){
          _masterScreen[col][row] = 0;
        }
      }
    }
  }
}




//**************** private methods*******************

/*
| - Description
|   - initialize screen array to zeros, initialize other
|   - registers and timer
|
*/
void masterScreen::init()
{
  for(int row = 0; row < 8; row++){
    for(int col = 0; col < 8; col++){
      _masterScreen[col][row] = 0;
    }
  }

  cur_col = 0;
  _shift_init();
  _screen_timer_init();
  cur_screen_obj_ptr =  this;
}


/*
| - Description
|   - Shifts byte into the shift registers
|
| - Parameter
|   - b   is a byte to be shifted in
|
*/
void masterScreen::_shift_byte_in(byte b)
{
  uint8_t i;
    for(i=0; i < 8; i++){

        if((b & _BV(i)))
            DATA_PORT |= _BV(DATA_PIN);
        else
            DATA_PORT &= ~_BV(DATA_PIN);

        CLK_PORT |= _BV(CLK_PIN);
        asm("nop");
        CLK_PORT &= ~_BV(CLK_PIN);
        asm("nop");
    }
}

/*
| - Description
|   - Shifts in the red, green, blue, and anode values into the registers
|   - in the correct order
|
| - Parameter
|   - red      Which red possitions are on for the cur_col
|   - green    Which green possitions are on for the cur_col
|   - blue     Which blue possitions are on for the cur_col
|   - anode    Which anode possitions are on for the cur_col
|
*/
void masterScreen::_shift_in(byte red, byte green, byte blue, byte anode)
{
       //[blue][red][green][anode]
    _shift_byte_in(anode);
    _shift_byte_in(green);
    _shift_byte_in(red);
    _shift_byte_in(blue);

    LATCH_PORT |= _BV(LATCH_PIN);
    asm("nop");
    LATCH_PORT &= ~_BV(LATCH_PIN);
}


void masterScreen::_shift_init()
{
    DDRD |= _BV(DATA_PIN);
    DDRD |= _BV(CLK_PIN);
    DDRD |= _BV(LATCH_PIN);
    DDRD |= _BV(RESET_PIN);
    DDRD |= _BV(EN_PIN);

     //make sure OE is low
    PORTD &= ~_BV(EN_PIN);
     //pull reset high
    PORTD |= _BV(RESET_PIN);
}


static void masterScreen::_screen_timer_init(){
    cli();
    TCCR1A = 0;
    TCCR1B = 0;
    // 1024 prescaler, 15625 Hz, 64us period
    TCCR1B |= _BV(CS12) | _BV(CS10);
    // clear on match
    TCCR1B |= _BV(WGM12);
    // little less than 1ms
    OCR1A = 15;
    // enable overflow interrupt
    TIMSK1 |= _BV(OCIE1A);
    sei();
}



void masterScreen::screen_display_col(byte col)
{
    char red   = 0x00;
    char green = 0x00;
    char blue  = 0x00;
    uint8_t col_value = 0x00;

    for(uint8_t i = 0; i < 8; i++){
        col_value = _masterScreen[col][i];
        red   |= ((col_value >> 0) & 0x01) << i;
        green |= ((col_value >> 1) & 0x01) << i;
        blue  |= ((col_value >> 2) & 0x01) << i;
    }
    _shift_in(~red, ~green, ~blue, _BV(col));
}


ISR(TIMER1_COMPA_vect)
{
    //_screen_add_sprites(screen_tmp);
    cur_screen_obj_ptr->screen_display_col(cur_col);
    cur_col++;
    if(cur_col > 7)
        cur_col = 0;
}
